"""ExoyOne CLI interface."""

import logging

import typer
from typer import Typer

from . import get, set

_LOGGER = logging.getLogger(__package__)

APP_NAME = "exoyone"

app: Typer = typer.Typer(no_args_is_help=True)
app.add_typer(get.app, name="get")
app.add_typer(set.app, name="set")


@app.callback()
def cli() -> None:
    """ExoyOne Command-Line Interface (CLI)."""


if __name__ == "__main__":
    """Run the ExoyOne CLI."""
    app()



"""ExoyOne state model."""

from __future__ import annotations

import logging
from dataclasses import dataclass, fields

from camel_converter import to_snake

from exoyone.models import mode_packs

_LOGGER = logging.getLogger(__package__)


@dataclass
class ExoyOneState:
    """ExoyOne state data model."""

    mdnsName: str
    type: int
    brightness: int
    currentModpack: int
    modeIndex: int
    speed: int
    hue: int
    saturation: int
    autoChange: bool
    musicSync: bool
    fadingOff: bool
    buttonEnabled: bool
    sceneGeneration: bool
    lockColorWheel: bool
    forceMusicSync: bool
    selectedPattern: int
    selectedRenderMode: int
    selectedColorMode: int
    selectedPalette: int
    userDefinedName: str
    cycleSpeed: int
    shutdownTimer: int
    direction: bool
    connectedToWiFi: bool
    firmwareVersion: str
    poweredByPowerbank: bool

    @property
    def friendly(self) -> dict[str, dict[str, int | str | tuple[int, int, int]]]:
        """Return fields in snake_case as a dictionary."""
        device_name = self.userDefinedName if self.userDefinedName else "Not Set"
        mode_pack = mode_packs.get_pack_name_from_index(self.currentModpack)
        effect = mode_packs.get_effect_name_from_index(
            self.currentModpack, self.modeIndex
        )

        def on_off(value: bool) -> str:
            """Return on if true, off if false."""
            return "On " if value else "Off "

        return {
            "Hardware Details": {
                "mDNS Name": f"{self.mdnsName}",
                "Firmware Version": self.firmwareVersion,
            },
            "Device Details": {
                "Device Name": device_name,
                "Power": on_off(self.fadingOff),
                "Powerbank Mode": on_off(self.poweredByPowerbank),
                "Shutdown Timer": self.shutdownTimer,
            },
            "Effect Details": {
                "Color": (self.hue, self.saturation, self.brightness),
                "Mode Pack": mode_pack,
                "Effect": effect,
                "Effect Speed": self.speed,
                "Effect Direction": self.direction,
            },
            "Effect Options": {
                "Music Sync": on_off(self.musicSync),
                "Scene Generation": on_off(self.sceneGeneration),
                "Mode Cycle": on_off(self.autoChange),
                "Mode Cycle Speed": self.cycleSpeed,
            },
        }

    def as_dict(self) -> dict[str, bool | int | str]:
        """Return fields in snake_case as a dictionary."""
        return {
            to_snake(dc_field.name): getattr(self, dc_field.name)
            for dc_field in fields(self)
        }



"""Exponential and predicate based backoff handlers."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    pass

_LOGGER = logging.getLogger(__package__)


class ExoyOneException(Exception):
    """ExoyOne exception."""


class ExoyOneTimeoutError(TimeoutError):
    """ExoyOne timeout error."""


class ExoyOneValueError(ValueError):
    """ExoyOne value error."""


class ExoyOneAttributeError(AttributeError):
    """ExoyOne attribute error."""


class ExoyOneAssertionError(AssertionError):
    """ExoyOne assertion error."""


class ExoyDevices(IntEnum):
    """Enumeration of Exoy devices."""

    INFINITY_OBJECT = 0
    HYPERCUBE = 1
    ULTRA_DENSE_HYPERCUBE = 2
    DODECAHEDRON = 3
    ULTRA_DENSE_DODECAHEDRON = 4
    MIRROR = 5
    ULTRA_DENSE_MIRROR = 6
    ICOSAHEDRON = 7
    ULTRA_DENSE_ICOSAHEDRON = 8
    TETRAHEDRON = 9
    ULTRA_DENSE_TETRAHEDRON = 10
    HEXAGON = 11
    ULTRA_DENSE_HEXAGON = 12
    SOUND_VISUALISER = 13
    ULTRA_DENSE_SOUND_VISUALISER = 14


class TruthyFalsyWords(IntEnum):
    """Truthy and falsy words."""

    TRUE = 1
    FALSE = 0
    ON = 1
    OFF = 0
    LEFT = 1
    RIGHT = 0
    CLOCKWISE = 1
    COUNTERCLOCKWISE = 0
    ANTICLOCKWISE = 0
    YES = 1
    NO = 0
    IN = 1
    OUT = 0
    UP = 1
    DOWN = 0
    FORWARD = 1
    BACKWARD = 0
    OPEN = 1
    CLOSE = 0


class ModePacks:
    """Helper class for mode packs and effects."""

    _mode_packs: ClassVar[list[str]] = [
        "Rave",
        "Party",
        "Chill",
        "Mood",
        "Genre",
        "Themes",
        "Nature",
        "Cosmic",
        "Background",
    ]

    _effects: ClassVar[list[list[str]]] = [
        [
            "Beats Burst",
            "Laser Lash",
            "Electric Ave",
            "Hypnosis",
            "Glow Groove",
            "Whirling Spectrum",
            "Rainbow Rumble",
        ],
        [
            "Dance Floor",
            "Rhythm Run",
            "Pop Party",
            "Funky Flash",
            "Jubulo",
            "Gloze",
            "Color Rush",
            "Confetti",
            "Rising Rhythm",
            "Pulse Peaks",
            "Beat Orbit",
        ],
        [
            "Soothing Sound",
            "Euphoria",
            "Dreamscape",
            "Serene Symphony",
            "Svelte",
            "Tranquilo",
        ],
        ["Mood Swing", "Emotional Express", "Feeling Flow", "Zenith", "Nimbus"],
        [
            "Rockin' Rhythms",
            "Jazz Live",
            "Hip Hop Hues",
            "Country Crescendo",
            "EDM Flow",
            "Hardstyle Beats",
        ],
        [
            "Cosmic Carnival",
            "Summer Breeze",
            "Desert Miracle",
            "Winter Wonderland",
            "Underwater Oasis",
            "Rainbow Ripple",
            "Error",
            "Matrix",
            "Beating Heart",
        ],
        [
            "Wildfire",
            "Thunderstorm",
            "Aurora",
            "Rainbow Reflection",
            "Bloom",
            "Celestial",
            "Ebb & Flow",
            "Majestic",
            "Whirlwind",
            "Lightning",
            "Snowstorm",
        ],
        [
            "Galaxy",
            "Cosmic Dust",
            "Solar Flare",
            "Sabre Fight",
            "Black Hole",
            "Meteor",
            "Astro",
            "Super Nova",
        ],
        [
            "Ambient Aura",
            "Twilight",
            "Dawn",
            "Custom Hue",
            "Color Carousel",
            "Shimmer",
            "Whisper",
        ],
    ]

    @property
    def mode_packs(self) -> list[str]:
        """Return a list of mode mode_pack names."""
        return self._mode_packs

    @property
    def pack_names(self) -> str:
        """Return mode_pack names as a comma-seperated string."""
        return f"[{'|'.join(name.lower() for name in self._mode_packs)}]"

    @property
    def effects(self) -> list[str]:
        """Return a list of effect names."""
        effect_names: list[str] = []
        for pack_effects in self._effects:
            effect_names.extend(pack_effects)
        return effect_names

    @property
    def effect_names(self) -> str:
        """Return effect names as a comma-seperated string."""
        return f"[{'|'.join(name.lower() for name in self.effects)}]"

    def get_effects_by_index(self, pack_index: int) -> list[str]:
        """Return a list of effects by mode mode_pack index."""
        return self._effects[pack_index]

    def get_pack_name_from_index(self, pack_index: int) -> str:
        """Return the name of a mode mode_pack given an index value."""
        return self.mode_packs[pack_index]

    def get_pack_index_from_name(self, name: str) -> int:
        """Return the index value of the named mode_pack."""
        try:
            return next(
                index
                for index, pack_name in enumerate(self._mode_packs)
                if pack_name.lower() == name.lower()
            )
        except StopIteration:
            return -1

    def get_effect_name_from_index(self, pack_index: int, effect_index: int) -> str:
        """Get effect name from current mode mode_pack and effect index."""
        active_effects = self.get_effects_by_index(pack_index)
        return active_effects[effect_index]

    def get_effect_index_from_name(self, pack_name: str, effect_name: str) -> int:
        """Return the index values of the mode_pack and effect named."""
        try:
            pack_index = self.get_pack_index_from_name(pack_name)
            if pack_index == -1:
                return -1
            pack_effects = self.get_effects_by_index(pack_index)
            return next(
                index
                for index, name in enumerate(pack_effects)
                if name.lower() == effect_name.lower()
            )
        except StopIteration:
            return -1

    def get_indices_from_effect_name(self, effect_name: str) -> tuple[int, int]:
        """Return the mode_pack index for the given effect name."""
        for index, _ in enumerate(self.mode_packs):
            effects = self.get_effects_by_index(pack_index=index)
            if effect_name in effects:
                return index, effects.index(effect_name)

        return -1, -1


mode_packs = ModePacks()



"""Set commands for ExoyOne CLI."""

from __future__ import annotations

import asyncio
import logging
from types import SimpleNamespace
from typing import Annotated

import rich
import typer
from click import Choice

from .. import ExoyOne
from .. import mode_packs as mp
from . import helpers as hp

_LOGGER = logging.getLogger(__package__)

app = typer.Typer(no_args_is_help=True)


@app.command("restart-in-ap-mode")
def restart_in_ap_mode(ctx: typer.Context) -> None:
    """Restart the ExoyOne to enable Wi-Fi access point."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    asyncio.run(exoyone.restart_in_ap_mode())


@app.command("color")
def color(
    ctx: typer.Context,
    hue: Annotated[
        int,
        typer.Argument(
            help="Hue value between 0 and 255.",
            metavar="[HUE]",
            show_default=False,
            min=0,
            max=255,
            clamp=True,
        ),
    ],
    saturation: Annotated[
        int,
        typer.Argument(
            help="Saturation value between 0 and 255.",
            metavar="[SAT]",
            show_default=False,
            min=0,
            max=255,
            clamp=True,
        ),
    ],
    brightness: Annotated[
        int,
        typer.Argument(
            help="Brightness value between 0 and 255.",
            metavar="[BRI]",
            show_default=False,
            min=0,
            max=255,
            clamp=True,
        ),
    ],
) -> None:
    """Set the color by changing the hue, saturation and brightness."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    hp.send_request_with_progress(
        exoyone,
        method_to_call="set_color",
        value_to_set=(hue, saturation, brightness),
        key_to_check="hue",
        description=f"Setting HSB to {hue}, {saturation}, {brightness}...",
        value_to_check=hue,
        value_to_bool=False,
    )


@app.command("shutdown-timer")
def shutdown_timer(
    ctx: typer.Context,
    minutes: Annotated[
        int,
        typer.Argument(
            help="Duration of shutdown timer from 5 to 480 minutes, 0 to disable.",
            metavar="[MINUTES]",
            min=0,
            max=480,
            show_choices=False,
            show_default=False,
        ),
    ],
) -> None:
    """Set the shutdown timer duration."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    if minutes == 0:
        description = "Disabling the shutdown timer..."
    else:
        description = f"Setting shutdown timer to {minutes} minutes..."
    hp.send_request_with_progress(
        exoyone,
        method_to_call="set_shutdown_timer",
        value_to_set=minutes,
        key_to_check="shutdownTimer",
        description=description,
        value_to_check=minutes * 60,
        value_to_bool=False,
    )


@app.command("effect")
def change_effect(
    ctx: typer.Context,
    name: Annotated[
        str,
        typer.Argument(
            click_type=Choice(choices=mp.effects, case_sensitive=False),
            help="Name of the effect.",
            metavar="[EFFECT]",
            case_sensitive=False,
            show_default=False,
            show_choices=False,
        ),
    ],
) -> None:
    """Change the active effect."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    pack_index, effect_index = mp.get_indices_from_effect_name(name)

    pack_name = mp.get_pack_name_from_index(pack_index)
    description = f"Changing mode pack to [red]{pack_name}[/red] and effect to [yellow]{name}[/yellow]..."  # noqa: E501

    hp.send_request_with_progress(
        exoyone,
        method_to_call="set_effect",
        value_to_set=(pack_index, effect_index),
        key_to_check="modexIndex",
        description=description,
        value_to_check=effect_index,
        value_to_bool=False,
    )


@app.command("effect-speed")
def effect_speed(
    ctx: typer.Context,
    speed: Annotated[
        int,
        typer.Argument(
            help="Effect speed", metavar="[SPEED]", show_default=False, min=0, max=250
        ),
    ],
) -> None:
    """Set the speed of the effect."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    hp.send_request_with_progress(
        exoyone,
        method_to_call="set_speed",
        value_to_set=speed,
        key_to_check="speed",
        description=f"Setting effect speed to {speed}...",
    )


@app.command("scene-generation")
def scene_generation(
    ctx: typer.Context,
    state: Annotated[
        hp.ToggleOptions,
        typer.Argument(
            help="Turn scene generation off or on.",
            metavar="[off|on]",
            case_sensitive=False,
            show_choices=False,
            show_default=False,
        ),
    ],
) -> None:
    """Turn scene generation off or on."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    value = str(state.value)
    hp.send_request_with_progress(
        exoyone,
        "toggle_scene_generation",
        value,
        "sceneGeneration",
        description=f"Turning scene generation {value}...",
    )


@app.command("mode-cycle")
def mode_cycle(
    ctx: typer.Context,
    state: Annotated[
        hp.ToggleOptions,
        typer.Argument(
            help="Turn mode cycle off or on.",
            metavar="[off|on]",
            case_sensitive=False,
            show_choices=False,
            show_default=False,
        ),
    ],
) -> None:
    """Turn mode cycle off or on."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    value = str(state.value)
    hp.send_request_with_progress(
        exoyone,
        "toggle_mode_cycle",
        value,
        "autoChange",
        description=f"Turning mode cycle {value}...",
    )


@app.command("cycle-speed")
def cycle_speed(
    ctx: typer.Context,
    seconds: Annotated[
        int,
        typer.Argument(
            help="Delay in seconds between cycles when mode cycle is on.",
            metavar="[SECONDS]",
            min=5,
            max=43200,
            clamp=True,
            show_default=False,
            show_choices=False,
        ),
    ],
) -> None:
    """Set the mode cycle duration in seconds."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    hp.send_request_with_progress(
        exoyone,
        method_to_call="set_cycle_speed",
        value_to_set=seconds,
        key_to_check="cycleSpeed",
        description=f"Setting mode cycle speed to {seconds} seconds...",
        value_to_check=seconds,
        value_to_bool=False,
    )


@app.command("music-sync")
def music_sync(
    ctx: typer.Context,
    state: Annotated[
        hp.ToggleOptions,
        typer.Argument(
            help="Turn music sync off or on.",
            metavar="[off|on]",
            case_sensitive=False,
            show_choices=False,
            show_default=False,
        ),
    ],
) -> None:
    """Turn music sync on or off."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    value = str(state.value)
    hp.send_request_with_progress(
        exoyone,
        "toggle_music_sync",
        value,
        "musicSync",
        description=f"Turning music sync {value}...",
    )


@app.command("power-state")
def power_state(
    ctx: typer.Context,
    state: Annotated[
        hp.ToggleOptions,
        typer.Argument(
            help="Switch the power on or off",
            metavar="[off|on]",
            case_sensitive=False,
            show_choices=False,
            show_default=False,
        ),
    ],
) -> None:
    """Turn power state on or off."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    value = str(state.value)
    hp.send_request_with_progress(
        exoyone,
        "toggle_power",
        value,
        "fadingOff",
        description=f"Turning power state {value}...",
    )


@app.command("powerbank-mode")
def powerbank_mode(
    ctx: typer.Context,
    state: Annotated[
        hp.ToggleOptions,
        typer.Argument(
            help="Switch powered by powerbank mode on or off",
            metavar="[off|on]",
            case_sensitive=False,
            show_choices=False,
            show_default=False,
        ),
    ],
) -> None:
    """Turn powered by powerbank mode on or off."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    value = str(state.value)
    hp.send_request_with_progress(
        exoyone,
        method_to_call="powered_by_powerbank",
        value_to_set=value,
        key_to_check="poweredByPowerbank",
        description=f"Turning powered by powerbank mode {value}...",
    )


@app.command("device-name")
def device_name(
    ctx: typer.Context,
    name: Annotated[
        str,
        typer.Argument(
            help="Custom name for the ExoyOne.", metavar="[NAME]", show_default=False
        ),
    ],
) -> None:
    """Set the device name of the ExoyOne."""
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    hp.send_request_with_progress(
        exoyone,
        method_to_call="set_name",
        value_to_set=name,
        key_to_check="userDefinedName",
        description=f"Setting device name to [yellow]{name}[/yellow]...",
        value_to_check=name,
        value_to_bool=False,
    )


@app.command("direction")
def direction(
    ctx: typer.Context,
    effect_direction: Annotated[
        hp.DirectionOptions,
        typer.Argument(
            help="Set the direction of the effect.",
            metavar="[this-way|that-way]",
            case_sensitive=False,
            show_choices=True,
            show_default=False,
        ),
    ],
) -> None:
    """
    Set a direction for the active effect.

    There are only two directions. The list of options is provided
    for entertainment purposes, as the actual values are 0 and 1.
    Feel free to mix left with down, right with forward and so on.
    """
    exoyone: ExoyOne = hp.get_exoyone(ctx.obj.host)
    value = str(effect_direction.value)

    to_set = -1
    if value.lower() in ["left", "in", "up", "forwards", "clockwise"]:
        to_set = 1
    if value.lower() in ["right", "out", "down", "backwards", "counterclockwise"]:
        to_set = 0

    hp.send_request_with_progress(
        exoyone,
        method_to_call="toggle_direction",
        value_to_set=to_set,
        key_to_check="direction",
        description=f"Setting effect direction to {value.title()}",
        value_to_bool=True,
    )


@app.callback()
def set(
    ctx: typer.Context,
    host: Annotated[
        str,
        typer.Option(
            "--host",
            "-H",
            help="Hostname or IP address of your ExoyOne.",
            case_sensitive=False,
            show_default=False,
            envvar="EXOYONE_HOST",
        ),
    ],
) -> None:  # pragma: no cover
    """
    Change things on your ExoyOne.

    This includes turning it on or off, changing the color, effect, and more.
    """
    if not host:
        rich.print("Missing required argument: '--host'")
        raise typer.Exit(1)

    ctx.obj = SimpleNamespace(host=host)



"""Get commands for exoyOneCLI."""

import logging
from types import SimpleNamespace
from typing import Annotated

import rich
import typer
from click import Choice
from rich import box
from rich.color import Color
from rich.columns import Columns
from rich.console import Console
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from ..models import mode_packs as mp
from . import helpers as hp

_LOGGER = logging.getLogger(__package__)
app = typer.Typer(no_args_is_help=True)


@app.command()
def effects(
    mode_pack: Annotated[
        str,
        typer.Option(
            click_type=Choice(choices=mp.mode_packs, case_sensitive=False),
            help="Name of a mode pack",
            show_choices=False,
            show_default=False,
        ),
    ]
    | None = None,
) -> None:
    """Get a list of effects for the specified mode pack or all mode packs."""
    trees: list[Tree] = []

    if mode_pack is not None:
        pack_index = mp.get_pack_index_from_name(mode_pack)
        if pack_index > -1:
            trees.append(Tree(f"[red]{mode_pack.title()}[/red]", guide_style="red"))
            mp_effects = mp.get_effects_by_index(pack_index)
            for effect in mp_effects:
                trees[0].add(effect)
        else:
            rich.print(f"Invalid mode pack: {mode_pack}")
            raise typer.Exit(2)
    else:
        for pack_index, pack_name in enumerate(mp.mode_packs):
            trees.append(
                Tree(
                    pack_name,
                    style=f"color({pack_index + 1})",
                )
            )

            for _, effect_name in enumerate(mp.get_effects_by_index(pack_index)):
                trees[pack_index].add(effect_name, style="white")

    columns = Columns(trees, equal=False, expand=False)
    rich.print(columns)


@app.command("power-state")
def power_state(ctx: typer.Context) -> None:
    """Get the power state (on/off)."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = "[green]ON[/]" if exoyone.state.fadingOff else "[red]OFF[/]"
    rich.print(f"Power state: {state}")


@app.command("music-sync")
def music_sync(ctx: typer.Context) -> None:
    """Get music sync state (on/off)."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = "[green]ON[/]" if exoyone.state.musicSync else "[red]OFF[/]"
    rich.print(f"Music sync: {state}")


@app.command("scene-generation")
def scene_generation(ctx: typer.Context) -> None:
    """Get scene generation state (on/off)."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = "[green]ON[/]" if exoyone.state.sceneGeneration else "[red]OFF[/]"
    rich.print(f"Scene generation: {state}")


@app.command("powerbank-mode")
def powerbank_mode(ctx: typer.Context) -> None:
    """Get powered by powerbank state (on/off)."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = "[green]ON[/]" if exoyone.state.poweredByPowerbank else "[red]OFF[/]"
    rich.print(f"Powered by powerbank mode: {state}")


@app.command("mode-cycle")
def mode_cycle(ctx: typer.Context) -> None:
    """Get mode cycle state (on/off)."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = "[green]ON[/]" if exoyone.state.autoChange else "[red]OFF[/]"
    rich.print(f"Mode cycle: {state}")


@app.command("cycle-speed")
def cycle_speed(ctx: typer.Context) -> None:
    """Get the current effect speed."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = str(exoyone.state.cycleSpeed)
    rich.print(f"Current mode cycle speed: [yellow]{state}[/]")


@app.command("direction")
def direction(ctx: typer.Context) -> None:
    """Get the effect direction (left/right)."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = "[green]LEFT[/]" if exoyone.state.direction else "[red]RIGHT[/]"
    rich.print(f"Effect direction: {state}")


@app.command("effect-speed")
def effect_speed(ctx: typer.Context) -> None:
    """Get the current effect speed."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = str(exoyone.state.speed)
    rich.print(f"Current effect speed: [yellow]{state}[/]")


@app.command("shutdown-timer")
def shutdown_timer(ctx: typer.Context) -> None:
    """Get the current shutdown timer."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    shutdown_timer_output = "Shutdown timer: "
    if exoyone.state.shutdownTimer > 0:
        minutes = int(exoyone.state.shutdownTimer / 60)
        shutdown_timer_output = f"{shutdown_timer_output}[yellow]{minutes} minutes[/]"
    else:
        shutdown_timer_output = f"{shutdown_timer_output}[red]Disabled[/]"
    rich.print(shutdown_timer_output)


@app.command("device-name")
def device_name(ctx: typer.Context) -> None:
    """Get the device name."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    dev_name = (
        f"[yellow]{exoyone.state.userDefinedName}[/]"
        if exoyone.state.userDefinedName
        else "[dim red]Unset[/]"
    )
    mdns_name = f"{exoyone.state.mdnsName}"
    rich.print(f"Device name: {dev_name}")
    rich.print(f"mDNS address: [blue]{mdns_name}[/]")


@app.command("color")
def color(ctx: typer.Context) -> None:
    """Get the current hue, saturation and brightness values."""
    exoyone = hp.get_exoyone(ctx.obj.host)

    console = Console()
    hue = exoyone.state.hue
    brightness = exoyone.state.brightness
    saturation = exoyone.state.saturation
    red, green, blue = hp.hsb_to_rgb(hue, saturation, brightness)
    color_obj: Color = Color.from_rgb(red, green, blue)

    console.print(
        Text()
        .append("RGB: ")
        .append(f"{red:>3}", style=f"{color_obj.name}")
        .append(", ")
        .append(f"{green:>3}", style=f"{color_obj.name}")
        .append(", ")
        .append(f"{blue:>3}", style=f"{color_obj.name}")
        .append(" [red, green, blue]", style="dim"),
    )
    console.print(
        Text()
        .append("HSB: ")
        .append(f"{hue:>3}", style=f"{color_obj.name}")
        .append(", ")
        .append(f"{saturation:>3}", style=f"{color_obj.name}")
        .append(", ")
        .append(f"{brightness:>3}", style=f"{color_obj.name}")
        .append(" [hue, saturation, brightness]", style="dim"),
    )
    console.print(
        Text()
        .append("CSS: ")
        .append(f"{color_obj.name}", style=f"{color_obj.name}")
        .append("       [HTML/CSS color]", style="dim"),
    )


@app.command("everything")
def everything(
    ctx: typer.Context,
) -> None:
    """Get the state and value of all the things."""
    exoyone = hp.get_exoyone(ctx.obj.host)
    state = exoyone.state

    console = Console()

    root = Table(
        box=None,
        highlight=True,
        show_footer=False,
        show_header=False,
    )

    root.add_column("Content", no_wrap=True)

    for section in state.friendly.keys():
        table = Table(
            box=box.SIMPLE_HEAD,
            highlight=True,
            show_footer=False,
            show_header=True,
        )

        table.add_column(f"{section}")
        table.add_column("Value")

        value: bool | int | str | tuple[int, int, int]
        for field, value in state.friendly[section].items():
            vt = Text()

            if field.lower() == "color" and isinstance(value, tuple):
                field = "Color: [[dim]Hue, Sat, Bri[/]]"
                (hue, sat, bri) = value
                red, green, blue = hp.hsb_to_rgb(hue, sat, bri)
                effect_color = Color.from_rgb(red, green, blue)

                hue_str = Text(f"({hue}, ", style=f"{effect_color.name}")
                sat_str = Text(f"{sat}, ", style=f"{effect_color.name}")
                bri_str = Text(f"{bri})", style=f"{effect_color.name}")

                vt = Text.assemble(hue_str, sat_str, bri_str)

            value_text = str(value)

            if field.lower() == "shutdown timer":
                value_text = "Disabled"
                if isinstance(value, int) and value > 0:
                    value_text = f"{int(value / 60)} minutes"

            elif field.lower() == "mode cycle speed":
                value_text = "Disabled"
                if state.autoChange is True:
                    value_text = f"{value} seconds"

            vt.append(Text(f"{value_text}", style="bold blue"))

            vt.highlight_words(["On ", "on "], style="bold green")
            vt.highlight_words(["Off ", "off "], style="bold red")
            vt.highlight_words(["Disabled"], style="dark_red")
            vt.highlight_words(mp.mode_packs, style="bold cyan")
            vt.highlight_words(mp.effects, style="bold magenta")
            cells = [field, vt]

            table.add_row(*cells)

        root.add_row(table, end_section=True)

    console.print(root)


@app.callback()
def get(
    ctx: typer.Context,
    host: Annotated[
        str,
        typer.Option(
            "--host",
            "-H",
            help="Hostname or IP address of your ExoyOne.",
            case_sensitive=False,
            show_default=False,
            envvar="EXOYONE_HOST",
        ),
    ],
) -> None:  # pragma: no cover
    """
    Get information from your ExoyOne.

    This includes the current state of the device, including power, color, effects,
    and more.
    """
    if not host:
        rich.print("Missing required argument: '--host'")
        raise typer.Exit(1)

    ctx.obj = SimpleNamespace(host=host)



"""Helper functions for the ExoyOne CLI."""

from __future__ import annotations

import asyncio
import colorsys
from enum import StrEnum

import rich
from rich.progress import Progress, SpinnerColumn, TextColumn

from .. import ExoyOne


def hsb_to_rgb(hue: int, saturation: int, brightness: int) -> tuple[int, int, int]:
    """Convert from HSB to RGB."""
    r, g, b = colorsys.hsv_to_rgb(hue / 255.0, saturation / 255.0, brightness / 255.0)
    return int(r * 255), int(g * 255), int(b * 255)


class DirectionOptions(StrEnum):
    """Direction options."""

    LEFT = "left"
    RIGHT = "right"
    IN = "in"
    OUT = "out"
    UP = "up"
    DOWN = "down"
    FORWARDS = "forwards"
    BACKWARDS = "backwards"
    CLOCKWISE = "clockwise"
    COUNTERCLOCKWISE = "counterclockwise"


class ToggleOptions(StrEnum):
    """Toggle options."""

    OFF = "off"
    ON = "on"


def send_request_with_progress(
    exoyone: ExoyOne,
    method_to_call: str,
    value_to_set: bool | int | str | tuple[int, int] | tuple[int, int, int],
    key_to_check: str,
    description: str = "Working...",
    value_to_check: bool | int | str | None = None,
    value_to_bool: bool = True,
) -> None:
    """Wrapper that calls an async toggle method with progress bar."""

    async def _with_progress() -> None:  # pragma: no cover
        """Async method with progress."""
        value = value_to_set

        if value_to_bool and (
            isinstance(value_to_set, bool)
            or isinstance(value_to_set, str)
            or isinstance(value_to_set, int)
        ):
            value = exoyone._bool_val(value_to_set)

        to_check = value_to_check if value_to_check is not None else value

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            task = progress.add_task(description, total=1)

            try:
                func = getattr(exoyone, method_to_call)
                await func(value)
                while not progress.finished:
                    if getattr(exoyone.state, key_to_check) == to_check:
                        progress.update(task, completed=1)
                    await asyncio.sleep(0.2)

            except AttributeError:
                return

    asyncio.run(_with_progress())
    rich.print(f"{description.removesuffix("...")}: [green]DONE[/green]")


def get_exoyone(host: str) -> ExoyOne:
    """Use asyncer to get the state of an ExoyOne instance."""
    exoyone = ExoyOne(host=host)

    async def _async_get() -> None:  # pragma: no cover
        """Connect to an ExoyOne device."""
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            task = progress.add_task(f"Connecting to {exoyone.host}...", total=1)
            await exoyone.async_get_data()

            while not progress.finished:
                if exoyone.state.currentModpack > -1:
                    progress.update(task, completed=1)
                await asyncio.sleep(0.1)

    asyncio.run(_async_get())
    return exoyone



# type: ignore

"""A mock ExoyOne device."""

import asyncio
import json
import logging
import random
from dataclasses import asdict, dataclass
from typing import Annotated

import asyncio_dgram
import typer
from rich.logging import RichHandler

FIELD_MAP = {
    "setBrightness": "brightness",
    "setModPack": "currentModpack",
    "setEffect": "modeIndex",
    "setSpeed": "speed",
    "setHue": "hue",
    "setSaturation": "saturation",
    "toggleModeCycle": ("autoChange", True),
    "toggleMusicSync": ("musicSync", True),
    "togglePower": ("fadingOff", True),
    "toggleSceneGeneration": ("sceneGeneration", True),
    "toggleDirection": ("direction", True),
    "setName": "userDefinedName",
    "setCycleSpeed": "cycleSpeed",
    "setShutdownTimer": "shutdownTimer",
    "poweredByPowerbank": ("poweredByPowerbank", True),
    "setPattern": "selectedPattern",
    "setRenderMode": "selectedRenderMode",
    "setColorMode": "selectedColorMode",
    "setPalette": "selectedPalette",
}

app = typer.Typer(no_args_is_help=True)


@dataclass
class MoxyOneState:
    """Fake MoxyOne state."""

    mdnsName: str
    type: int = 4
    brightness: int = 0
    currentModpack: int = 0
    modeIndex: int = 0
    speed: int = 10
    hue: int = 0
    saturation: int = 255
    autoChange: bool = False
    musicSync: bool = False
    fadingOff: bool = False
    buttonEnabled: bool = True
    sceneGeneration: bool = False
    lockColorWheel: bool = False
    forceMusicSync: bool = False
    selectedPattern: int = 1
    selectedRenderMode: int = 1
    selectedColorMode: int = 1
    selectedPalette: int = 1
    userDefinedName: str = ""
    cycleSpeed: int = 20
    shutdownTimer: int = 0
    direction: bool = True
    connectedToWiFi: bool = True
    firmwareVersion: str = "2.1"
    poweredByPowerbank: bool = False

    def as_dict(self) -> dict:
        """Return as dict."""
        values = asdict(self)
        values["shutdownTimer"] = values["shutdownTimer"] * 60
        return values


class MoxyOne:
    """A mock ExoyOne device."""

    def __init__(self, logger: logging.Logger, delay: bool = False) -> None:
        """Initialize MoxyOne."""
        self._state = MoxyOneState(
            mdnsName=f"exoyone{random.randint(12345, 67890):05d}"
        )
        self._stream: asyncio_dgram.aio.DatagramServer | None = None
        self._lock = asyncio.Lock()
        self._logger = logger
        self._timeout: float = 0.1
        self._delay = delay

    async def serve(self) -> None:
        """Start responding to requests."""
        self._stream = await asyncio_dgram.bind(("127.0.0.1", 8888))
        self._logger.info("MoxyOne started serving on %s", self._stream.sockname)
        self._logger.debug("Current status: %s", self._state.as_dict())

        while True:
            try:
                data, remote_addr = await self._stream.recv()
                request = json.loads(data.decode())
                self._logger.debug("Received %s from %s", request, remote_addr)

                if request.get("getData", False) is False:
                    for key, value in request.items():
                        self._logger.info("Handling %s = %s", key, value)
                        if key in FIELD_MAP:
                            param = FIELD_MAP[key]
                            if len(param) == 2:
                                value = bool(value)
                                param = param[0]
                            elif key == "setShutdownTimer":
                                value = value.get("minutes") + (value.get("hours") * 60)
                                param = param

                            elif key == "cycleSpeed":
                                value = value.get("minutes") + (value.get("hours") * 60)
                                param = param
                            elif key == "connectToWiFi" or key == "restartInApMode":
                                continue
                            self._logger.info(
                                "Changing %s from %s to %s",
                                param,
                                getattr(self._state, param),
                                value,
                            )
                            setattr(self._state, f"{param}", value)
                        else:
                            self._logger.info("Ignoring %s", key)
                            continue
                elif request.get("getData", False) == 1:
                    if self._delay:
                        await asyncio.sleep(0.6)
                    self._logger.info(
                        "Sending %s to %s", self._state.as_dict(), remote_addr
                    )
                    await self._stream.send(
                        json.dumps(self._state.as_dict()).encode(), remote_addr
                    )

            except asyncio_dgram.aio.TransportClosed:
                self._logger.error("Transport closed.")
            except Exception as e:
                self._logger.error("Error: %s", e, exc_info=e)


@app.command()
def serve(ctx: typer.Context) -> None:
    """Start the server."""
    loop = asyncio.get_event_loop()
    logger = ctx.obj.get("logger")
    moxyone = MoxyOne(logger)

    task = loop.create_task(moxyone.serve())

    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        task.cancel()
    except asyncio.CancelledError:
        pass


@app.callback()
def main(
    ctx: typer.Context,
    debug: Annotated[bool, typer.Option("--debug")] = False,
    delay: Annotated[bool, typer.Option("--delay")] = False,
) -> None:
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler()],
    )

    _LOGGER = logging.getLogger(__package__)

    if debug:
        _LOGGER.setLevel(level=logging.DEBUG)

    ctx.ensure_object(dict)
    ctx.obj["logger"] = _LOGGER

    if delay:
        ctx.obj["delay"] = True


if __name__ == "__main__":
    app()



__version__ = "0.1.10"

from .exoyone import ExoyOne
from .models import (
    ExoyOneAssertionError,
    ExoyOneException,
    ExoyOneTimeoutError,
    ExoyOneValueError,
    mode_packs,
)
from .models import ModePacks as ExoyOneModePacks
from .state import ExoyOneState

__all__ = [
    "ExoyOne",
    "ExoyOneAssertionError",
    "ExoyOneException",
    "ExoyOneModePacks",
    "ExoyOneState",
    "ExoyOneTimeoutError",
    "ExoyOneValueError",
    "mode_packs",
]



"""ExoyOne integration."""

from __future__ import annotations

import asyncio
import json
import logging
from collections.abc import Mapping
from typing import ClassVar

import backoff
from asyncio_dgram.aio import DatagramClient, connect

from . import __version__
from .models import (
    ExoyDevices,
    ExoyOneTimeoutError,
    ExoyOneValueError,
    ModePacks,
    TruthyFalsyWords,
)
from .state import ExoyOneState

_LOGGER = logging.getLogger(__package__)
_GET_DATA = b'{"getData": 1}'

pyexoyone_version = __version__


class ExoyOne:
    """Representation of an ExoyOne Light."""

    TIMEOUT: ClassVar[float] = 3.0

    def __init__(self, host: str, port: int = 8888) -> None:
        """Initialize the ExoyOne library."""
        self._host = host
        self._port = port
        self._state: ExoyOneState
        self._mp = ModePacks()

    @property
    def host(self) -> str:
        """Return the hostname or IP address used to connect."""
        return self._host

    @property
    def name(self) -> str:
        """Return either the user-defined name or mDNS name."""
        name = self._state.mdnsName
        if len(self._state.userDefinedName) > 0:
            name = self._state.userDefinedName
        return name

    @property
    def state(self) -> ExoyOneState:
        """Returns the in-memory state of the ExoyOne."""
        return self._state

    @property
    def device_type(self) -> str:
        """Return the device type."""
        device_type = ExoyDevices(self._state.type).name
        return device_type.replace("_", " ").title()

    @staticmethod
    def _bool_val(value: bool | int | str) -> int:
        """Return 1 for truthy state or 0 for falsy state."""
        if isinstance(value, int):
            return value

        if isinstance(value, str) and value.upper() in TruthyFalsyWords.__members__:
            return TruthyFalsyWords[value.upper()].value

        raise ExoyOneValueError(f"Invalid value: {value}")

    @backoff.on_exception(
        backoff.expo,
        ExoyOneTimeoutError,
        max_tries=3,
        logger=_LOGGER,
        backoff_log_level=logging.DEBUG,
    )
    async def async_set_data(
        self, request: Mapping[str, Mapping[str, int | str] | bool | int | str]
    ) -> None:
        """Update a setting on the ExoyOne, then get the latest state data."""
        encoded_update = json.dumps(request).encode("utf-8")
        try:
            stream: DatagramClient = await connect((self._host, self._port))
            await asyncio.wait_for(stream.send(encoded_update), timeout=self.TIMEOUT)
        except TimeoutError as exc:
            raise ExoyOneTimeoutError() from exc
        finally:
            stream.close()
            del stream
            await self.async_get_data()

    @backoff.on_exception(
        backoff.expo,
        ExoyOneTimeoutError,
        max_tries=3,
        logger=_LOGGER,
        backoff_log_level=logging.DEBUG,
    )
    async def async_get_data(self) -> None:
        """Update the in-memory state using data from the ExoyOne."""
        try:
            stream: DatagramClient = await connect((self._host, self._port))
            await asyncio.wait_for(stream.send(b'{"getData": 1}'), timeout=self.TIMEOUT)
            reply, _ = await asyncio.wait_for(stream.recv(), timeout=self.TIMEOUT)
            data = json.loads(reply.decode("utf-8"))
            self._state = ExoyOneState(**data)
        except TimeoutError as exc:
            raise ExoyOneTimeoutError() from exc
        finally:
            stream.close()
            del stream

    def get_active_pack_name(self) -> str:
        """Return the name of the currently active modpack."""
        return self._mp.get_pack_name_from_index(self._state.currentModpack)

    def get_active_effect(self) -> str:
        """Return the name of the currently active effect."""
        return self._mp.get_effect_name_from_index(
            pack_index=self._state.currentModpack,
            effect_index=self._state.modeIndex,
        )

    async def async_get_state(self) -> ExoyOneState:
        """Return the current state as an ExoyOne object."""
        await self.async_get_data()
        return self._state

    async def restart_in_ap_mode(self) -> None:
        """Restart the device in AP mode."""
        await self.async_set_data({"restartInApMode": True})

    async def toggle_power(self, state: bool | int | str) -> None:
        """Toggle power."""
        request = {"togglePower": self._bool_val(state)}
        await self.async_set_data(request)

    async def toggle_direction(self, state: bool | int | str) -> None:
        """Toggle direction."""
        request = {"toggleDirection": self._bool_val(state)}
        await self.async_set_data(request)

    async def _async_set_hsbs(self, key: str, value: int) -> None:
        """Helper method to set hue, saturation, brightness and speed."""
        value = max(0, min(value, 255))
        await self.async_set_data({key: value})

    async def set_color(self, hsb: tuple[int, int, int]) -> None:
        """Set the hue, saturation and brightness in a single call."""
        hue = max(0, min(hsb[0], 255))
        saturation = max(0, min(hsb[1], 255))
        brightness = max(0, min(hsb[2], 255))
        await self.async_set_data(
            {
                "setHue": hue,
                "setSaturation": saturation,
                "setBrightness": brightness,
            }
        )

    async def set_hue(self, hue: int) -> None:
        """Set the hue."""
        await self._async_set_hsbs(key="setHue", value=hue)

    async def set_saturation(self, saturation: int) -> None:
        """Set the saturation."""
        await self._async_set_hsbs(key="setSaturation", value=saturation)

    async def set_brightness(self, brightness: int) -> None:
        """Set the brightness."""
        await self._async_set_hsbs(key="setBrightness", value=brightness)

    async def set_speed(self, speed: int) -> None:
        """Set the speed."""
        await self._async_set_hsbs(key="setSpeed", value=speed)

    async def toggle_mode_cycle(self, state: bool | int | str) -> None:
        """Toggle automatic mode change after set interval."""
        request = {"toggleModeCycle": self._bool_val(state)}
        await self.async_set_data(request)

    async def toggle_music_sync(self, state: bool | int | str) -> None:
        """Toggle music sync."""
        request = {"toggleMusicSync": self._bool_val(state)}
        await self.async_set_data(request)

    async def set_cycle_speed(self, cycle_speed: int) -> None:
        """Set cycle speed in seconds."""
        await self.async_set_data({"setCycleSpeed": cycle_speed})

    async def set_effect(self, effect: str | tuple[int, int]) -> None:
        """Set mode mode_pack and effect."""
        if isinstance(effect, str):
            pack_index, effect_index = self._mp.get_indices_from_effect_name(effect)
        else:
            pack_index, effect_index = effect

        if pack_index > -1 and effect_index > -1:
            await self.async_set_data(
                {
                    "setModPack": pack_index,
                    "setEffect": effect_index,
                }
            )

    async def toggle_scene_generation(self, state: bool | int | str) -> None:
        """Toggle scene generation."""
        request = {"toggleSceneGeneration": self._bool_val(state)}
        await self.async_set_data(request)

    async def set_name(self, name: str) -> None:
        """Set a new user-friendly name."""
        if len(name) > 39:
            name = name[:39]
            _LOGGER.warning("Name truncated to maximum allowed length: %s", name[:39])
        await self.async_set_data({"setName": name})

    async def set_shutdown_timer(self, minutes: int) -> None:
        """Set a shutdown timer of at least 5 and at most 480 minutes."""
        minutes = int(minutes)

        if minutes == 0:
            _LOGGER.debug("Disabling shutdown timer.")

        elif 1 < minutes < 5:
            _LOGGER.debug("Shutdown timer set to minimum duration of 5 minutes.")
            minutes = 5

        if minutes > 480:
            _LOGGER.debug("Shutdown timer set to maximum duration of 8 hours.")
            minutes = 480

        hours: int = 0
        if minutes > 60:
            hours = minutes // 60
            minutes = minutes % 60

        await self.async_set_data(
            {"setShutdownTimer": {"hours": hours, "minutes": minutes}}
        )

    async def connect_to_wifi(self, ssid: str, password: str) -> None:
        """Set new Wi-Fi connection credentials."""
        if len(ssid) == 0 or len(password) == 0:
            raise ExoyOneValueError(
                "Both SSID and password must be set to change WiFi credentials"
            )
        if len(ssid) > 31:
            raise ExoyOneValueError(
                "SSID is longer than maximum allowed length of 31 characters"
            )
        if len(password) > 31:
            raise ExoyOneValueError(
                "Password is longer than the maximum allowed length of 31 characters"
            )

        await self.async_set_data(
            {"connectToWifi": {"ssid": ssid, "password": password}}
        )

    async def powered_by_powerbank(self, state: bool | int | str) -> None:
        """Toggle power consumption reduction when running on batteries."""
        request = {"poweredByPowerbank": self._bool_val(state)}
        await self.async_set_data(request)

    async def set_pattern(self, value: int) -> None:
        """Use async_set_data to change the selectedPattern value."""
        await self.async_set_data({"setPattern": value})

    async def set_palette(self, value: int) -> None:
        """Use async_set_data to change the selectedPalette value."""
        await self.async_set_data({"setPalette": value})

    async def set_render_mode(self, value: int) -> None:
        """Use async_set_data to change the selectedRenderMode value."""
        await self.async_set_data({"setRenderMode": value})

    async def set_color_mode(self, value: int) -> None:
        """Use async_set_data to change the selectedColorMode value."""
        await self.async_set_data({"setColorMode": value})
